# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function
import yaml
import sys
import os
import tempfile
import traceback
import subprocess
import array
import cairo
import ftplib
import random
import string
import datetime
from pw_utils import PasswordGenerator
try:
    from gi import require_version
except ImportError:
    print("ERROR: pygobject is not installed.")
    raise
try:
    require_version("Gtk", "3.0")
except ValueError:
    print("ERROR: gtk3 development files are not installed.")
    raise
except AttributeError:
    print("pygobject version too old.")
    raise
from gi.repository import Gtk
from gi.repository import Gdk
# from gi.repository import GLib
from gi.repository import GObject
from os.path import abspath
from os.path import dirname
from os.path import isfile
from os.path import join
from pkg_resources import resource_exists
from pkg_resources import resource_filename
from logging import getLogger
from logging import basicConfig
from logging import INFO
from six import iteritems
from six import StringIO
from printingdemo import PrintingApp
from jscript_utils import shift
try:
    from synapselabels import __version__
    assert __version__
except ImportError:
    __version__ = "UNKNOWN"


log = getLogger('synapselabels')
dnd_targets = [('text/plain', Gtk.TargetFlags.OTHER_WIDGET, 0), ]
default_unit = 'm'


def get_path(name):
    """
    Returns the absolute path name of a resource file
    File in the current directory overrides file in package
    """
    path = None
    if isfile(name):
        path = abspath(name)
    elif resource_exists(__name__, name):
        res_filename = resource_filename(__name__, name)
        res_path = abspath(join(dirname(__file__), res_filename))
        if isfile(res_path):
            path = res_path
    if path:
        log.info("located path: %s", path)
    else:
        log.error("could not locate path for: %s", name)
    return path


label_config_file = get_path('labelconfig.yaml')
glade_file = get_path('labelsgui.glade')
src1 = get_path('names.txt')
src2 = get_path('places.txt')


def pil2cairo(im):
    """Transform a PIL Image into a Cairo ImageSurface."""

    assert sys.byteorder == 'little', 'We don\'t support big endian'
    if im.mode != 'RGBA':
        im = im.convert('RGBA')

    s = im.tobytes('raw', 'BGRA')
    a = array.array('B', s)
    dest = cairo.ImageSurface(cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx = cairo.Context(dest)
    non_premult_src_wo_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_RGB24, im.size[0], im.size[1])
    non_premult_src_alpha = cairo.ImageSurface.create_for_data(
        a, cairo.FORMAT_ARGB32, im.size[0], im.size[1])
    ctx.set_source_surface(non_premult_src_wo_alpha)
    ctx.mask_surface(non_premult_src_alpha)
    return dest


def get_selected_data(treeview):
        # print("print_selected_label")
        (model, pathlist) = treeview.get_selection().get_selected_rows()
        for path in pathlist:
            tree_iter = model.get_iter(path)
            for index, column in enumerate(
                    treeview.get_columns()):
                return model.get_value(tree_iter, index)


def on_drag_data_get(treeview, context, selection, info, timestamp):
    """ Extract data from the source of the DnD operation.

    Serialize iterators of selected tasks in format
    <iter>,<iter>,...,<iter> and set it as parameter of DND """

    selected_data = get_selected_data(treeview)
    selection.set(selection.get_target(), 0, selected_data)


def set_treeview_columns(treeview, columns):
    for i in range(len(columns)):
        # cellrenderer to render the text
        cell = Gtk.CellRendererText(editable=False)
        # cell.set_property("editable", True)
        # the text in the first column should be in boldface
        # if i == 0:
        #    cell.props.weight_set = True
        #    cell.props.weight = Pango.Weight.BOLD
        # the column is created
        col = Gtk.TreeViewColumn(
            columns[i], cell, text=i)
        # and it is appended to the treeview
        treeview.append_column(col)


def set_treeview_from_clipboard(treeview, with_header=True):
    set_treeview_from_text_data(
        treeview,
        Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).wait_for_text(),
        with_header
        )


def set_treeview_from_text_data(treeview, text_data, with_header=True):
    listmodel = None
    if text_data is not None:
        # print("clipboard text:", repr(text_data))
        for line in text_data.splitlines():
            # print("line:", repr(line))
            texts = line.split('\t')
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                args = [GObject.TYPE_STRING] * len(texts)
                # print(args)
                listmodel = Gtk.ListStore(*args)
                # listmodel = Gtk.ListStore(str, str, str)
                if with_header:
                    columns = texts
                else:
                    columns = [
                        "Column{}".format(index) for
                        index in
                        range(0, len(texts))
                        ]
                    # print("appending:", texts)
                    listmodel.append(texts)
            else:
                # print("appending:", texts)
                listmodel.append(texts)
        if listmodel is not None:
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, columns)
    else:
        print("No text on the clipboard.")


def set_treeview_from_texts(treeview, texts):
    # treeview.set_headers_visible(False)
    listmodel = None
    if texts is not None:
        # print("texts:", repr(texts))
        for line in texts:
            # print("line:", repr(line))
            if listmodel is None:
                for column in treeview.get_columns():
                    treeview.remove_column(column)
                listmodel = Gtk.ListStore(GObject.TYPE_STRING)
                # print("appending:", line)
                listmodel.append([line])
            else:
                # print("appending:", line)
                listmodel.append([line])
        if listmodel is not None:
            # print("setting model:", listmodel)
            treeview.set_model(listmodel)
            set_treeview_columns(treeview, ['Params'])


def fill_combo(combo, data):
    store = Gtk.ListStore(GObject.TYPE_STRING)
    for line in data:
        store.append([line])
    combo.set_model(store)
    combo.set_active(0)
    # And here's the new stuff:
    cell = Gtk.CellRendererText()
    combo.pack_start(cell, True)
    combo.add_attribute(cell, "text", 0)


def get_treeview_column(treeview, x_pos):
    for column in treeview.get_columns():
        if column.get_x_offset() <= x_pos:
            if x_pos <= (column.get_x_offset() + column.get_width()):
                return column


def drag_data_received(
        widget, context, x_pos, y_pos, selection, info, timestamp):
    data = selection.get_data()
    print('drag_data_received data:', data)
    if data:
        column = get_treeview_column(widget, x_pos)
        if column:
            print(
                "old column title:", repr(column.get_title()),
                "new column title:", repr(data)
                )
            column.set_title(data)


class LabelConfig(object):

    def __init__(self, config_file_path):
        with open(config_file_path, 'r') as stream:
            yaml_data = yaml.load(stream)
        # print('yaml_data:', repr(yaml_data))
        # yaml.dump(yaml_data, sys.stdout, default_flow_style=False)
        if 'label types' in yaml_data:
            self.label_types = yaml_data['label types']
            # print('label types:', repr(self.label_types))
        else:
            raise ValueError('"label types" is not found in config file')
        if 'printers' in yaml_data:
            self.printers = yaml_data['printers']
            # print('printers:', repr(printers))
        else:
            raise ValueError('"printers" is not found in config file')
        if 'models' in yaml_data:
            self.models = yaml_data['models']
            # print('models:', repr(self.models))
            self.model_names = []
            for model in self.models:
                for model_name in model:
                    self.model_names.append(model_name)
        else:
            raise ValueError('"models" is not found in config file')
        if 'labels' in yaml_data:
            self.labels = yaml_data['labels']
        else:
            raise ValueError('"labels" is not found in config file')


class LabelsGui(LabelConfig):
    """
    This is the E20 GUI
    """
    def __init__(self):
        """
        Build GUI
        """
        super(LabelsGui, self).__init__(label_config_file)
        self.model = None
        self.dpi = 300.0
        # Build GUI from Glade file
        self.builder = Gtk.Builder()
        self.builder.add_from_file(glade_file)
        # Get objects
        go = self.builder.get_object
        self.window = go('applicationwindow1')
        self.window.set_title('Synapse Labels - v{}'.format(__version__))
        self.textview = go('textview')
        self.drawingarea = go('drawingarea')
        self.drawingarea.connect("draw", self.on_draw)
        self.treeview_data_table = go('treeviewDataTable')
        self.treeview_data_table.get_selection().connect(
            "changed",
            self.data_tree_selection_changed,
            "dummy"
            )
        self.treeview_data_table.enable_model_drag_dest(
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_data_table.connect(
            'drag-data-received',
            drag_data_received,
            )
        # self.combobox = go('comboboxtext1')
        # self.mac_num = go('labelMacNum')
        self.entry_date_code = go('entryDateCode')
        self.entry_records = go('entryRecords')
        self.entry_records.connect(
            "changed",
            self.entry_records_changed
            )
        self.button_today = go('buttonToday')
        self.entry_copies = go('entryCopies')
        self.button_today.connect(
            "clicked",
            self.button_today_clicked,
            "today button data"
            )
        self.button_generate = go('buttonGenerate')
        self.button_generate.connect(
            "clicked",
            self.button_generate_clicked,
            )
        self.button1 = go('button1')
        self.button1.connect(
            "clicked",
            self.button_1_clicked,
            )
        self.button2 = go('button2')
        self.button2.connect(
            "clicked",
            self.button_2_clicked,
            )
        self.button3 = go('button3')
        self.button3.connect(
            "clicked",
            self.button_3_clicked,
            )
        self.button4 = go('button4')
        self.button4.connect(
            "clicked",
            self.button_4_clicked,
            )
        self.button5 = go('button5')
        self.button5.connect(
            "clicked",
            self.button_5_clicked,
            )
        self.label_number_of_records = go('labelNumberOfRecords')
        self.buttonPasteWithHeader = go('buttonPasteWithHeader')
        self.buttonPasteWithHeader.connect(
            "clicked",
            self.button_paste_with_header_clicked,
            "cool button 2"
            )
        self.buttonPrintSelectedLabels = go('buttonPrintSelectedLabels')
        self.buttonPrintSelectedLabels.connect(
            "clicked",
            self.print_selected_labels,
            )
        self.buttonPrintEveryLabel = go('buttonPrintEveryLabel')
        self.buttonPrintEveryLabel.connect(
            "clicked",
            self.print_every_label,
            )

        self.buttonSheetSelectedLabels = go('buttonSheetSelectedLabels')
        self.buttonSheetSelectedLabels.connect(
            "clicked",
            self.sheet_selected_labels,
            )
        self.buttonSheetEveryLabel = go('buttonSheetEveryLabel')
        self.buttonSheetEveryLabel.connect(
            "clicked",
            self.sheet_every_label,
            )

        self.cb_printers = go('comboboxPrinters')
        self.cb_models = go('comboboxModels')
        self.cb_models.connect(
            "changed",
            self.cb_models_selection_changed,
            "dummy"
            )

        self.treeview_params = go('treeviewParams')
        self.treeview_params.get_selection().connect(
            "changed",
            self.params_tree_selection_changed,
            "dummy"
            )
        self.treeview_params.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            dnd_targets,
            Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE
            )
        self.treeview_params.connect('drag_data_get', on_drag_data_get)

        fill_combo(self.cb_printers, self.printers.keys())
        fill_combo(self.cb_models, self.model_names)
        self.set_date_code_to_today()
        # Connect signals
        self.builder.connect_signals(self)
        self.cb_models_selection_changed(self.cb_models)
        # Configure interface
        self.window.connect('delete-event', lambda x, y: Gtk.main_quit())
        # GObject.timeout_add(1000, self.update_mac_pool)
        # Everything is ready
        self.window.show()
        self.password_generator = PasswordGenerator(src1, src2)

    def print_selected_labels(self, widget, data=None):
        self.print_labels(self.get_selected_labels())

    def sheet_selected_labels(self, widget, data=None):
        self.sheet_labels(self.get_selected_labels())

    def get_selected_labels(self):
        # print("print_selected_label")
        selection = self.treeview_data_table.get_selection()
        # print('selection:', selection)
        (model, pathlist) = selection.get_selected_rows()
        labels = []
        for path in pathlist:
            # print('path:', path)
            tree_iter = model.get_iter(path)
            # print('tree_iter:', tree_iter)
            label_data = {}
            for param in self.get_parameters():
                value = None
                for index, column in enumerate(
                        self.treeview_data_table.get_columns()):
                    if column.get_title() == param:
                        value = model.get_value(
                            tree_iter,
                            index
                            )
                        break
                # print('value:', value)
                if value is None:
                    if param == 'date code':
                        entry_date_code_text = self.entry_date_code.get_text()
                        if len(entry_date_code_text) == 4:
                            value = entry_date_code_text
                if value:
                    label_data[param] = value
                else:
                    print(repr(param), 'parameter is missing')
                    return
            labels.append(label_data)
        return labels

    def get_label_from_row(self, row):
        label_data = {}
        for param in self.get_parameters():
            value = None
            for index, column in enumerate(
                    self.treeview_data_table.get_columns()):
                if column.get_title() == param:
                    value = row[index]
                    break
            # print('value:', value)
            if value:
                label_data[param] = value
            else:
                print(repr(param), 'parameter is missing')
                return
        return label_data

    def print_every_label(self, widget, data=None):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        self.print_labels(labels)

    def sheet_every_label(self, widget, data=None):
        labels = []
        for row in self.treeview_data_table.get_model():
            label_data = self.get_label_from_row(row)
            if label_data:
                labels.append(label_data)
        self.sheet_labels(labels)

    def button_1_clicked(self, widget, data=None):
        self.entry_copies.set_text('1')

    def button_2_clicked(self, widget, data=None):
        self.entry_copies.set_text('2')

    def button_3_clicked(self, widget, data=None):
        self.entry_copies.set_text('3')

    def button_4_clicked(self, widget, data=None):
        self.entry_copies.set_text('4')

    def button_5_clicked(self, widget, data=None):
        self.entry_copies.set_text('5')

    def get_number_of_records(self):
        return len(self.treeview_data_table.get_model())

    def adjust_number_records(self):
        self.label_number_of_records.set_text(
            str(
                self.get_number_of_records()
                )
            )

    def button_paste_clicked(self, widget, data=None):
        # print("button_paste_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=False)
        self.adjust_number_records()

    def button_paste_with_header_clicked(self, widget, data=None):
        # print("button_paste_with_header_clicked - %s" % data)
        set_treeview_from_clipboard(self.treeview_data_table, with_header=True)
        self.adjust_number_records()

    def data_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        csr = widget.count_selected_rows()
        # print("count_selected_rows:", csr)
        self.buttonPrintSelectedLabels.set_sensitive(csr > 0)
        self.buttonSheetSelectedLabels.set_sensitive(csr > 0)
        self.update_label_preview()

    def update_label_preview(self):
        selected_labels = self.get_selected_labels()
        if selected_labels is not None and len(selected_labels) == 1:
            self.size_x, self.size_y = self.get_label_size()
            self.size_x /= 25.4
            self.size_y /= 25.4
            self.selected_label = selected_labels[0]
            # print('preview_label:', self.selected_label)
            self.selected_label_file = self.get_label_file(self.selected_label)
            # Gnm.textview_set_text(
            #     self.textview.set_text,
            #     self.selected_label_file
            #     )
            self.textview.get_buffer().set_text(self.selected_label_file)
            # self.preview_label_file(self.get_label_file(label))
            self.drawingarea.set_size_request(
                300.0 * self.size_x,
                300.0 * self.size_y
                )
            self.drawingarea.queue_draw()
            # self.window.reshow_with_initial_size()
        else:
            self.textview.get_buffer().set_text('')
            self.drawingarea.set_size_request(-1, -1)
            self.selected_label = None
            self.label_file = None
            # self.window.reshow_with_initial_size()

    # def preview_label_file(self, label_file):
    #     print('preview_label_file', label_file)
    #     for line in label_file.splitlines():
    #         if line.startswith('T'):
    #             print ("preview", line)
    #             control, text = line.split(';', 1)
    #             control = control[2:]
    #             pos_x, pos_y, rotate, font, size = control.split(',')
    #             print ("text", pos_x, pos_y, rotate, font, size, text)

    def dots(self, size, unit):
        if size.startswith('pt'):
            size = float(size[2:],) / 72.0
        else:
            size = float(size)
            if unit == 'm':
                size = size / 25.6  # convert mm to inch
            elif unit != 'i':
                raise ValueError(
                    'invalid unit: %s' % repr(unit)
                    )
        return self.dpi * size

    def on_draw(self, wid, cr):
        if not hasattr(self, 'size_x'):
            return
        if not hasattr(self, 'size_y'):
            return
        cr.set_font_size(300.0 * 6.0 / 72)
        # frame it!
        cr.set_line_width(1)
        cr.rectangle(
            0,
            0,
            300 * self.size_x,
            300 * self.size_y,
            )
        cr.set_source_rgb(255, 255, 255)
        cr.fill()
        unit = default_unit
        for line in self.selected_label_file.splitlines():
            if line.startswith('m'):
                unit = line[2:]  # m or i
                if unit == 'i':
                    # print('unit is inch')
                    pass
                elif unit == 'm':
                    # print('unit is mm')
                    pass
                else:
                    print('unsupported line:', line)
            elif line.startswith('J'):
                if len(line) != 1:
                    print('unsupported line:', line)
            elif line.startswith('S'):
                # size TBD
                pass
            elif line.startswith('O'):
                # TBD
                pass
            elif line.startswith('A'):
                # TBD
                pass
            elif line.startswith('T'):
                # print("preview", line)
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, font, size = control.split(',')
                # print("text", pos_x, pos_y, rotate, font, size, text)
                cr.move_to(self.dots(pos_x, unit), self.dots(pos_y, unit))
                size = self.dots(size, unit)
                cr.set_font_size(size)
                if font == '3':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_NORMAL,
                        )
                elif font == '5':
                    cr.select_font_face(
                        "Swiss 721",
                        cairo.FONT_SLANT_NORMAL,
                        cairo.FONT_WEIGHT_BOLD,
                        )
                cr.set_source_rgb(0, 0, 0)
                cr.show_text(text)
            elif line.startswith('G'):
                # print("preview", line)
                # G 0.05,0.64,0;R:1.9,0.33,0.0066666
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate = control.split(',')
                function, function_params = text.split(':', 1)
                if function == 'R':
                    width, height, thickness = function_params.split(',', 2)
                    cr.set_line_width(self.dots(thickness, unit))
                    cr.rectangle(
                        self.dots(pos_x, unit),
                        self.dots(pos_y, unit),
                        self.dots(width, unit),
                        self.dots(height, unit),
                        )
                    cr.set_source_rgb(0, 0, 0)
                    cr.stroke()
                else:
                    print('unsupported line:', line)
            elif line.startswith('B'):
                # print("preview", line)
                # B 1.39,0.08,0,QRCODE+EL{error level}+MODEL2,.0133333;{qrcode}
                control, text = line.split(';', 1)
                control = control[2:]
                pos_x, pos_y, rotate, barcode, size = control.split(',', 4)
                barcode_opts = barcode.split('+')
                if barcode_opts[0] == 'QRCODE':
                    qr_model = 1
                    error_correction = 'L'
                    for opt in barcode_opts:
                        if opt == 'QRCODE':
                            pass
                        elif opt == 'MODEL1':
                            qr_model = 1
                        elif opt == 'MODEL2':
                            qr_model = 2
                        elif opt == 'ELL':
                            error_correction = 'L'
                        elif opt == 'ELM':
                            error_correction = 'M'
                        elif opt == 'ELQ':
                            error_correction = 'Q'
                        elif opt == 'ELH':
                            error_correction = 'H'
                        else:
                            print('unsupported qrcode option:', opt)
                    box_size = int(0.5 + self.dots(size, unit))
                    if qr_model != 2:
                        print('unsupported QR Code model:', qr_model)
                    try:
                        with tempfile.NamedTemporaryFile(
                            suffix='.png',
                            prefix='qr',
                            delete=False
                                ) as qr_file:
                            qr_file_name = qr_file.name
                        subprocess.check_call(
                            [
                                'qrencode',
                                '-o', qr_file_name,
                                '-l', error_correction,
                                '-m', '0',
                                '-s', str(box_size),
                                text,
                                ]
                            )
                        surface = cairo.ImageSurface.create_from_png(
                            qr_file_name
                            )
                    finally:
                        try:
                            if os.path.isfile(qr_file_name):
                                os.remove(qr_file_name)
                                # print('removed temporary file:', qr_file_name)
                        except:
                            print('cannot remove temporary file:', qr_file_name)
                            traceback.print_exc(file=sys.stdout)
                    cr.set_source_surface(
                        surface,
                        self.dots(pos_x, unit),
                        self.dots(pos_y, unit)
                        )
                    cr.paint()
                    # print('cr', dir(cr))
                else:
                    print('unsupported line:', line)
            else:
                print('unsupported line:', line)

    def params_tree_selection_changed(self, widget, data=None):
        # print("selection_changed - %s" % data)
        # csr = widget.count_selected_rows()
        # print("params_tree_selection_changed count_selected_rows:", csr)
        pass

    def get_label_name(self):
        # print('get_label_name')
        for model in self.models:
            # print('model:', model)
            for model_name, label_name in iteritems(model):
                # print('model name:', model_name, 'label name:', label_name)
                if model_name == self.model:
                    # print('label_name:', label_name)
                    return label_name

    def get_label(self):
        # print('get_label')
        label = self.labels[self.get_label_name()]
        # print('label:', label)
        return label

    def get_parameters(self):
        # print('get_parameters')
        try:
            parameters = self.get_label()['parameters']
            # print('parameters:', parameters)
            return parameters
        except:
            # silent fail, because it is possible to have a label without param
            return []

    def get_label_type(self):
        return self.label_types[self.get_label()['label type']]

    def get_label_size(self):
        label_type = self.get_label_type()
        return label_type['width'], label_type['height']

    def cb_models_selection_changed(self, widget, data=None):
        # print("cb_models_selection_changed - %s" % data)
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            self.model = widget.get_model()[tree_iter][0]
            # print("Selected model:", self.model)
            set_treeview_from_texts(
                self.treeview_params,
                self.get_parameters()
                )
        self.update_label_preview()

    def button_today_clicked(self, widget, data=None):
        # print("button_today_clicked - %s" % data)
        self.set_date_code_to_today()

    def set_date_code_to_today(self):
        today = datetime.date.today()
        todays_week = today.strftime("%W")
        todays_year = today.strftime("%Y")
        todays_date_code = str(todays_year[-2:]) + str(todays_week)
        self.entry_date_code.set_text(todays_date_code)

    def get_label_file(self, label_params, x_pos=None):
        for param in self.get_parameters():
            if param not in label_params:
                raise ValueError('label parameter %s is missing' % repr(param))

        label_format_params = {'model': self.model}
        label_name = self.get_label_name()
        if label_name == "#02 with meid":
            label_format_params['qrcode'] = (
                "#02:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address'] + ":" +
                label_params['snap mac address'] + ":" +
                label_params['wifi mac address'] + ":" +
                label_params['modem meid']
                )
            label_format_params['error level'] = 'H'
        elif label_name == "#02 without meid":
            label_format_params['qrcode'] = (
                "#02:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address'] + ":" +
                label_params['snap mac address'] + ":" +
                label_params['wifi mac address']
                )
            label_format_params['error level'] = 'H'
        elif label_name == "#08":
            label_format_params['qrcode'] = (
                "#08:" +
                self.model + ":" +
                label_params['date code'] + ":" +
                label_params['ethernet mac address last 6'][-6:] + ":" +
                label_params['snap node address'][-6:] + ":" +
                label_params['wifi mac address'] + ":" +
                label_params['modem meid'] + ":" +
                label_params['sim iccid'] + ":"
                )
            label_format_params['error level'] = (
                'Q' if len(label_format_params['qrcode']) >= 80 else 'H'
                )
        elif label_name == "#05":
            label_format_params['error level'] = 'L'
        label_format_params.update(label_params)
        # print('label_format_params:', label_format_params)
        retval = ''
        for item in self.get_label()['items']:
            if x_pos is not None:
                retval += shift(
                    item.format(**label_format_params),
                    x_pos
                    ) + '\n'
            else:
                retval += item.format(**label_format_params) + '\n'
        return retval

    def print_label_file_to(self, label_file, printer_config):
        # create random file name
        filename = ''.join(random.choice(string.digits) for _ in range(9))
        # print("random file name:", filename)
        if 'ftp' in printer_config:
            ftp_config = printer_config['ftp']
            ftp_connection = None
            try:
                ftp_connection = ftplib.FTP(
                    ftp_config['ip address']
                    )
                ftp_connection.login(
                    ftp_config['user name'],
                    ftp_config['password']
                    )
                ftp_connection.cwd("execute")
                ftp_connection.storlines(
                    'STOR ' + filename,
                    StringIO(label_file)
                    )
            except:
                print("FTP Error")
                traceback.print_exc(file=sys.stdout)
            finally:
                if ftp_connection is not None:
                    ftp_connection.close()

    def get_selected_printer_config(self):
        tree_iter = self.cb_printers.get_active_iter()
        if tree_iter is not None:
            selected_printer_name = self.cb_printers.get_model()[tree_iter][0]
            return self.printers[selected_printer_name]

    def print_label_file(self, label_file):
        # print('print_label_file:\n"""\n%s"""' % label_file)
        selected_printer_config = self.get_selected_printer_config()
        if selected_printer_config is not None:
            self.print_label_file_to(label_file, selected_printer_config)

    def sheet_labels(self, labels, copies=1):
        app = PrintingApp()
        app.run()

    def print_labels(self, labels, copies=None):
        # print('printing model:', self.model, 'labels:', labels)
        if copies is None:
            copies = int(self.entry_copies.get_text())
        x_pos = 0
        lpr = self.get_label_type()['lpr']
        total_width = self.get_label_type()['width']
        if lpr > 1:
            total_width += (lpr - 1) * self.get_label_type()['hrepeat']
        # print('lpr:', lpr)
        label_file = None
        if labels is not None:
            for label in labels:
                for copy in range(copies):
                    current_label_file = self.get_label_file(
                        label,
                        (x_pos * self.get_label_type()['hrepeat'])
                        if x_pos else None
                        )
                    if label_file is None:
                        label_file = 'm m\n'
                        label_file += 'J\n'
                        label_file += (
                            'S l1;0,0,' +
                            str(self.get_label_type()['height']) + ',' +
                            str(self.get_label_type()['vrepeat']) + ',' +
                            str(total_width) +
                            '\n'
                            )
                        label_file += 'O R\n'
                    label_file += current_label_file
                    x_pos += 1
                    if x_pos >= lpr:
                        label_file += 'A 1\n'
                        self.print_label_file(label_file)
                        label_file = None
                        x_pos = 0
        if label_file is not None:
            label_file += 'A 1\n'
            self.print_label_file(label_file)
            label_file = None
            x_pos = 0

    def button_generate_clicked(self, widget, data=None):
        # print("button_generate_clicked - %s" % data)
        set_treeview_from_text_data(
            self.treeview_data_table,
            self.generate_records(),
            with_header=True,
            )
        self.adjust_number_records()

    def entry_records_changed(self, widget, data=None):
        self.button_generate.set_sensitive(False)
        records = int(self.entry_records.get_text())
        # print('records', records)
        self.button_generate.set_sensitive(records > 0)

    def generate_records(self):
        try:
            records = int(self.entry_records.get_text())
        except ValueError:
            records = 0
        # print('records', records)
        parameters = self.get_parameters()
        retval = '\t'.join(parameters) + '\n'
        for record in range(records):
            retval += '\t'.join(
                [self.generate_parameter(parameter) for parameter in parameters]
                ) + '\n'
        return retval

    def generate_parameter(self, parameter):
        if parameter in ['user password', 'wifi password']:
            return self.password_generator.generate_password()
        else:
            raise ValueError(
                "cannot generation of parameter: %s" %
                repr(parameter)
                )


def synapse_labels():
    basicConfig(level=INFO)
    gui = LabelsGui()
    assert gui
    Gtk.main()


if __name__ == "__main__":
    synapse_labels()
